/*
 * Copyright (c) 2015-2022 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Mon Aug 24 2015
 */
#include "acpom.h"

#include <sys/time.h>
#include <aio.h>
#include <err.h>
#include <unistd.h>

#include <cassert>
#include <cmath>
#include <cstdio>

#include "pom_c_types.h"
#include "codels.h"


/* --- Task io ---------------------------------------------------------- */


/** Codel pom_io_start of task io.
 *
 * Triggered by pom_start.
 * Yields to pom_read.
 * Throws pom_e_sys.
 */
genom_event
pom_io_start(pom_ids *ids, const genom_context self)
{
  struct timeval now;
  pom_snapshot s;

  ids->context = new pom_context_s;
  ids->history_length = 2; /* 2s by default */
  ids->fperiodms = pom_period_ms; /* 1ms by default */

  /* set initial state */
  gettimeofday(&now, NULL);

  ids->context->ts.sec = now.tv_sec;
  ids->context->ts.nsec = now.tv_usec * 1000;

  ids->context->state = ukf::state_s::Identity();
  ids->context->cov = ukf::state_s::dcov_s::Identity();

  ids->pmodel = ukf::state_s::pmodel;
  ids->max_jerk = ukf::state_s::max_dadt;
  ids->max_dw = ukf::state_s::max_dwdt;

  ids->magdir.x = ukf::measure_s::magdir(0);
  ids->magdir.y = ukf::measure_s::magdir(1);
  ids->magdir.z = ukf::measure_s::magdir(2);

  ids->offset.p.x = ids->offset.p.y = ids->offset.p.z = 0.;
  ids->offset.q.qx = ids->offset.q.qy = ids->offset.q.qz = 0.;
  ids->offset.q.qw = 1.;

  /* init logging */
  try {
    ids->log_state = new pom_log_s;
    ids->log_measurements = new pom_log_s;
  } catch(...) {
    errno = ENOMEM;
    return pom_e_sys_error("init", self);
  }
  ids->log_state->req.aio_fildes = -1;
  ids->log_state->req.aio_offset = 0;
  ids->log_state->req.aio_buf = ids->log_state->buffer;
  ids->log_state->req.aio_nbytes = 0;
  ids->log_state->req.aio_reqprio = 0;
  ids->log_state->req.aio_sigevent.sigev_notify = SIGEV_NONE;
  ids->log_state->req.aio_lio_opcode = LIO_NOP;
  ids->log_state->pending = false;
  ids->log_state->skipped = false;

  ids->log_measurements->req.aio_fildes = -1;
  ids->log_measurements->req.aio_offset = 0;
  ids->log_measurements->req.aio_buf = ids->log_measurements->buffer;
  ids->log_measurements->req.aio_nbytes = 0;
  ids->log_measurements->req.aio_reqprio = 0;
  ids->log_measurements->req.aio_sigevent.sigev_notify = SIGEV_NONE;
  ids->log_measurements->req.aio_lio_opcode = LIO_NOP;
  ids->log_state->pending = false;
  ids->log_state->skipped = false;

  return pom_read;
}


/** Codel pom_io_read of task io.
 *
 * Triggered by pom_read.
 * Yields to pom_pause_read, pom_insert.
 * Throws pom_e_sys.
 */
genom_event
pom_io_read(const pom_measure *measure, sequence_pom_portinfo_s *ports,
            sequence_pom_measurement_s *measurements,
            const genom_context self)
{
  const or_pose_estimator_state *m;
  size_t i;

  for(i = 0; i < ports->_length; i++) {
    if (!ports->_buffer[i].name[0]) continue;

    /* get current port data */
    if (measure->read(ports->_buffer[i].name, self) != genom_ok)
      continue;
    m = measure->data(ports->_buffer[i].name, self);
    if (!m) continue;

    /* do not consider already read data */
    if (m->ts.nsec == ports->_buffer[i].last.nsec &&
        m->ts.sec == ports->_buffer[i].last.sec) continue;
    ports->_buffer[i].last = m->ts;

    /* record data */
    assert(measurements->_length < measurements->_maximum);
    measurements->_buffer[measurements->_length].data = *m;
    measurements->_buffer[measurements->_length].offset =
      ports->_buffer[i].offset;
    measurements->_buffer[measurements->_length].id = i;
    measurements->_length++;
  }

  return measurements->_length ? pom_insert : pom_pause_read;
}


/** Codel pom_io_insert of task io.
 *
 * Triggered by pom_insert.
 * Yields to pom_pause_read.
 * Throws pom_e_sys.
 */
genom_event
pom_io_insert(sequence_pom_measurement_s *measurements,
              pom_context_s **context, double history_length,
              pom_log_s **log_measurements, const genom_context self)
{
  pom_measurement_s *m;
  pom_snapshot s;
  struct timeval now;
  ssize_t avail;
  double delay;
  size_t i;
  char *l;
  int n;

  gettimeofday(&now, NULL);

  /* check if logging is requested and possible */
  l = NULL;
  avail = 0;
  if ((*log_measurements)->req.aio_fildes >= 0) {
    /* check for previous log request status */
    if ((*log_measurements)->pending) {
      if (aio_error(&(*log_measurements)->req) != EINPROGRESS) {
        /* fulfilled */
        (*log_measurements)->pending = false;

        /* close log in case of I/O error */
        if (aio_return(&(*log_measurements)->req) <= 0) {
          warn("log_measurements");
          close((*log_measurements)->req.aio_fildes);
          (*log_measurements)->req.aio_fildes = -1;
        }
      } else {
        /* not fulfilled, next log will be skipped */
        (*log_measurements)->skipped = true;
        (*log_measurements)->missed++;
      }
    }

    /* prepare a new log buffer if no request is pending */
    if ((*log_measurements)->req.aio_fildes >= 0 &&
        !(*log_measurements)->pending) {
      (*log_measurements)->total++;
      l = (*log_measurements)->buffer;
      avail = sizeof((*log_measurements)->buffer);

      /* marks skipped entries with an empty line */
      if ((*log_measurements)->skipped) {
        n = snprintf(l, avail, "\n");
        l += n; avail -= n; assert(avail > 0 && "log buffer too small");
        (*log_measurements)->skipped = false;
      }
    }
  }

  /* insert new measurements with an empty state */
  s.fused = false;
  s.state = ukf::state_s::Identity();
  s.cov = ukf::state_s::dcov_s::Identity();

  for(i = 0; i < measurements->_length; i++) {
    m = &measurements->_buffer[i];
    s.measurement = *m;
    (*context)->history.insert(s);

    /* build log buffer */
    if (l) {
      if (!m->data.pos._present) {
        m->data.pos._value.x = m->data.pos._value.y = m->data.pos._value.z = 0.;
      }
      if (!m->data.att._present) {
        m->data.att._value.qw =
          m->data.att._value.qx =
          m->data.att._value.qy =
          m->data.att._value.qz = 0.;
      }
      if (!m->data.vel._present) {
        m->data.vel._value.vx = m->data.vel._value.vy =
        m->data.vel._value.vz = 0.;
      }
      if (!m->data.avel._present) {
        m->data.avel._value.wx = m->data.avel._value.wy =
        m->data.avel._value.wz = 0.;
      }

      if (!m->data.acc._present) {
        m->data.acc._value.ax = m->data.acc._value.ay =
        m->data.acc._value.az = 0.;
      }

      delay = now.tv_sec - m->data.ts.sec +
        1e-9 * (1e3 * now.tv_usec - m->data.ts.nsec);

      double qw = m->offset.q.qw;
      double qx = m->offset.q.qx, qy = m->offset.q.qy, qz = m->offset.q.qz;
      double roll  = atan2(2 * (qw*qx + qy*qz), 1 - 2 * (qx*qx + qy*qy));
      double pitch = asin(2 * (qw*qy - qz*qx));
      double yaw   = atan2(2 * (qw*qz + qx*qy), 1 - 2 * (qy*qy + qz*qz));

      n = snprintf(
        l, avail, pom_log_m_line_fmt,
        m->id, delay, m->data.ts.sec, m->data.ts.nsec,
        m->data.intrinsic,
        m->data.pos._present, m->data.att._present,
        m->data.vel._present, m->data.avel._present,
        m->data.acc._present,
        m->data.pos._value.x, m->data.pos._value.y, m->data.pos._value.z,
        m->data.att._value.qw,
        m->data.att._value.qx, m->data.att._value.qy, m->data.att._value.qz,
        m->data.vel._value.vx, m->data.vel._value.vy, m->data.vel._value.vz,
        m->data.avel._value.wx, m->data.avel._value.wy, m->data.avel._value.wz,
        m->data.acc._value.ax, m->data.acc._value.ay, m->data.acc._value.az,
        m->offset.p.x, m->offset.p.y, m->offset.p.z,
        roll, pitch, yaw);
      l += n; avail -= n; assert(avail > 0 && "log buffer too small");

#define pom_log_cov(m, x)                                               \
      do {                                                              \
        size_t j;                                                       \
                                                                        \
        for(j = 0;                                                      \
            j < sizeof((m)->data. x ## _cov._value.cov)/sizeof(double); \
            j++) {                                                      \
          double c = (m)->data. x ## _cov._present ?                    \
                     (m)->data. x ## _cov._value.cov[j] : 0.;           \
          n = snprintf(l, avail, pom_logfmt, c);                        \
          l += n; avail -= n;                                           \
          assert(avail > 0 && "log buffer too small");                  \
        }                                                               \
      } while(0)

      pom_log_cov(m, pos);
      pom_log_cov(m, att);
      pom_log_cov(m, att_pos);

      pom_log_cov(m, vel);
      pom_log_cov(m, avel);

      pom_log_cov(m, acc);

#undef pom_log_cov

      n = snprintf(l, avail, "\n");
      l += n; avail -= n; assert(avail > 0 && "log buffer too small");
    }
  }
  measurements->_length = 0;

  /* truncate history */
  s.measurement.data.ts.sec = now.tv_sec - (time_t)history_length;
  n = (history_length - (time_t)history_length)*1e6;
  if (now.tv_usec >= n)
    s.measurement.data.ts.nsec = (now.tv_usec - n) * 1000;
  else {
    s.measurement.data.ts.sec--;
    s.measurement.data.ts.nsec = (now.tv_usec + 1000000 - n) * 1000;
  }
  pom_history_s::iterator last = (*context)->history.upper_bound(s);

  if (last != (*context)->history.end()) {
    for(;!last->fused && last != (*context)->history.end(); ++last)
      warnx("history too short");

    if (last != (*context)->history.end())
      (*context)->history.erase(++last, (*context)->history.end());
  }

  /* issue log request */
  if (l) {
    (*log_measurements)->req.aio_nbytes =
      sizeof((*log_measurements)->buffer) - avail;
    if (aio_write(&(*log_measurements)->req)) {
      warn("log_measurements");
      close((*log_measurements)->req.aio_fildes);
      (*log_measurements)->req.aio_fildes = -1;
    } else
      (*log_measurements)->pending = true;
  }

  return pom_pause_read;
}


/** Codel pom_io_stop of task io.
 *
 * Triggered by pom_stop.
 * Yields to pom_ether.
 * Throws pom_e_sys.
 */
genom_event
pom_io_stop(pom_ids *ids, const genom_context self)
{
  if (ids->context) delete ids->context;
  ids->context = NULL;

  pom_log_stop(&ids->log_state, &ids->log_measurements, self);
  delete ids->log_state;
  delete ids->log_measurements;
  ids->log_state = NULL;
  ids->log_measurements = NULL;

  return pom_ether;
}


/* --- Activity add_measurement ----------------------------------------- */

/** Codel pom_add_measurement of activity add_measurement.
 *
 * Triggered by pom_start.
 * Yields to pom_ether.
 * Throws pom_e_sys.
 */
genom_event
pom_add_measurement(const char port[128], double x, double y, double z,
                    double roll, double pitch, double yaw,
                    sequence_pom_portinfo_s *ports,
                    sequence_pom_measurement_s *measurements,
                    const pom_measure *measure,
                    const genom_context self)
{
  size_t i;

  if (measure->read(port, self) != genom_ok) {
    errno = ENOENT;
    return pom_e_sys_error("add", self);
  }

  for(i = 0; i < ports->_length; i++) {
    if (!strcmp(ports->_buffer[i].name, port)) break;
  }

  if (i >= ports->_length) {
    if (genom_sequence_reserve(measurements, i + 1))
      return pom_e_sys_error("add", self);
    if (genom_sequence_reserve(ports, i + 1))
      return pom_e_sys_error("add", self);

    ports->_length = i + 1;
  }

  strncpy(
    ports->_buffer[i].name, port, sizeof(ports->_buffer->name)-1);
  ports->_buffer[i].name[sizeof(ports->_buffer->name)-1] = 0;

  ports->_buffer[i].offset.p.x = x;
  ports->_buffer[i].offset.p.y = y;
  ports->_buffer[i].offset.p.z = z;

  double cr = cos(roll/2.), cp = cos(pitch/2.), cy = cos(yaw/2.);
  double sr = sin(roll/2.), sp = sin(pitch/2.), sy = sin(yaw/2.);
  ports->_buffer[i].offset.q.qw = cr * cp * cy + sr * sp * sy;
  ports->_buffer[i].offset.q.qx = sr * cp * cy - cr * sp * sy;
  ports->_buffer[i].offset.q.qy = cr * sp * cy + sr * cp * sy;
  ports->_buffer[i].offset.q.qz = cr * cp * sy - sr * sp * cy;

  ports->_buffer[i].last.sec = 0;
  ports->_buffer[i].last.nsec = 0;

  return pom_ether;
}


/* --- Activity replay -------------------------------------------------- */

static FILE *replayf;
static double replayoff;

/** Codel pom_replay_start of activity replay.
 *
 * Triggered by pom_start.
 * Yields to pom_read, pom_ether.
 * Throws pom_e_sys.
 */
genom_event
pom_replay_start(const char path[128], uint32_t *lineno,
                 const genom_context self)
{
  char header[1024];
  char *s;

  replayf = fopen(path, "r");
  if (!replayf) return pom_e_sys_error(path, self);

  /* discard header */
  s = fgets(header, sizeof(header), replayf);
  if (!s) return pom_stop;

  replayoff = nan("");
  *lineno = 2;
  return pom_read;
}

/** Codel pom_replay_read of activity replay.
 *
 * Triggered by pom_read.
 * Yields to pom_insert, pom_stop.
 * Throws pom_e_sys.
 */
genom_event
pom_replay_read(sequence_pom_measurement_s *measurements,
                uint32_t *lineno, const genom_context self)
{
  pom_measurement_s m;
  struct timeval now;
  struct timespec dt, rdt;
  int posp, attp, velp, avelp, accp;
  double roll, pitch, yaw;
  double t;
  unsigned int id;
  int i, c;

  do {
    /* get next measurement */
    c = fscanf(
      replayf,
      pom_scan_m_line_fmt,
      &id, &m.data.ts.sec, &m.data.ts.nsec,
      &i, &posp, &attp, &velp, &avelp, &accp,
      &m.data.pos._value.x, &m.data.pos._value.y, &m.data.pos._value.z,
      &m.data.att._value.qw,
      &m.data.att._value.qx, &m.data.att._value.qy, &m.data.att._value.qz,
      &m.data.vel._value.vx, &m.data.vel._value.vy, &m.data.vel._value.vz,
      &m.data.avel._value.wx, &m.data.avel._value.wy, &m.data.avel._value.wz,
      &m.data.acc._value.ax, &m.data.acc._value.ay, &m.data.acc._value.az,
      &m.offset.p.x, &m.offset.p.y, &m.offset.p.z,
      &roll, &pitch, &yaw);
    if (c == EOF) return pom_stop;
    else if (c != 32) {
      warnx("bad state format, line %d", *lineno);
      errno = EBADF;
      return pom_e_sys_error("replay", self);
    }

    double cr = cos(roll/2.), cp = cos(pitch/2.), cy = cos(yaw/2.);
    double sr = sin(roll/2.), sp = sin(pitch/2.), sy = sin(yaw/2.);
    m.offset.q.qw = cr * cp * cy + sr * sp * sy;
    m.offset.q.qx = sr * cp * cy - cr * sp * sy;
    m.offset.q.qy = cr * sp * cy + sr * cp * sy;
    m.offset.q.qz = cr * cp * sy - sr * sp * cy;

    m.id = id;
    m.data.intrinsic = i;

    m.data.pos._present = m.data.pos_cov._present = posp;
    m.data.att._present = m.data.att_cov._present = attp;
    m.data.att_pos_cov._present = posp && attp;

    m.data.vel._present = m.data.vel_cov._present = velp;
    m.data.avel._present = m.data.avel_cov._present = avelp;

    m.data.acc._present = m.data.acc_cov._present = accp;
    m.data.aacc._present = m.data.aacc_cov._present = false;

#define pom_scan_cov(f, m, x)                                           \
    do {                                                                \
      size_t j;                                                         \
                                                                        \
      for(j = 0;                                                        \
          j < sizeof((m).data. x ## _cov._value.cov)/sizeof(double);    \
          j++) {                                                        \
        if (!fscanf(f, "%lf", &(m).data. x ## _cov._value.cov[j])) {    \
          warnx("bad covariance format, line %d", *lineno);             \
          errno = EBADF;                                                \
          return pom_e_sys_error("replay", self);                       \
        }                                                               \
      }                                                                 \
    } while(0)

    pom_scan_cov(replayf, m, pos);
    pom_scan_cov(replayf, m, att);
    pom_scan_cov(replayf, m, att_pos);

    pom_scan_cov(replayf, m, vel);
    pom_scan_cov(replayf, m, avel);

    pom_scan_cov(replayf, m, acc);

#undef pom_scan_cov

    if (!fscanf(replayf, "\n"))
      /* tolerate missing newline, but consume it if present */;
    (*lineno)++;

    /* get time offset */
    gettimeofday(&now, NULL);
    if (std::isnan(replayoff)) {
      replayoff = now.tv_sec + 1e-6*now.tv_usec
                  - m.data.ts.sec - 1e-9*m.data.ts.nsec;
    }

    t = m.data.ts.sec + 1e-9*m.data.ts.nsec + replayoff;
    m.data.ts.sec = floor(t);
    m.data.ts.nsec = 1e9*(t - m.data.ts.sec);

    /* get deadline */
    t = m.data.ts.sec + 1e-9*m.data.ts.nsec - now.tv_sec - 1e-6*now.tv_usec;

    /* record data */
    if (measurements->_length >= measurements->_maximum)
      if (genom_sequence_reserve(measurements, 1 + measurements->_maximum))
        return pom_e_sys_error("replay", self);

    measurements->_buffer[measurements->_length] = m;
    measurements->_length++;
  } while(t < -1e-3);

  /* wait until deadline */
  if (t > 1e-3) {
    dt.tv_sec = floor(t);
    dt.tv_nsec = 1e9*(t - dt.tv_sec);
    while(nanosleep(&dt, &rdt)) {
      if (errno != EINTR) break;
      dt = rdt;
    }
  }

  return pom_insert;
}

/** Codel pom_io_insert of activity replay.
 *
 * Triggered by pom_insert.
 * Yields to pom_pause_read.
 * Throws pom_e_sys.
 */
/* already defined in task io */


/** Codel pom_replay_stop of activity replay.
 *
 * Triggered by pom_stop.
 * Yields to pom_ether.
 * Throws pom_e_sys.
 */
genom_event
pom_replay_stop(const genom_context self)
{
  if (replayf) {
    fclose(replayf);
    replayf = NULL;
  }

  return pom_ether;
}
